/*
 * sample.c - random sampling of STS(19)
 *
 * Copyright (c) 2008, Petteri Kaski, Patric Ostergard, Olli Pottonen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* The function random() is an X/Open System Interfaces Extension.
 * For C99 compatibility, we need the sixth issue. */
#define _XOPEN_SOURCE 600

#include <time.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include "compression.h"

#define STS_PER_FILE 125000000
#define STS_COUNT    11084874829
#define FILE_COUNT   89
#define FILENAME_LEN 9 /* Length of "sts19-XX" including terminating '\0' */

static file_r *files[FILE_COUNT];
static int     files_initialized = 0;

static void            init_rnd();
static unsigned long long int rnd();

static void            init_files(const char *dir);
const int            * get_sts(long long idx);

/* We need good pseudo-random unsigned long long int. When possible, we use
 * random() since it is usually much better than rand(). In either case we
 * need to combine several random int's to get a random long long int.
 */
#if defined(__WIN32__)
static void init_rnd() {
    srand((unsigned int)time(NULL));
}

static unsigned long long int rnd() {
    return (((unsigned long long) rand())) ^
        (((unsigned long long) rand()) << 16) ^
        (((unsigned long long) rand()) << 32) ^
        (((unsigned long long) rand()) << 48);
}
#else
static void init_rnd() {
    srandom((unsigned int)time(NULL));
}

static unsigned long long int rnd() {
    return (((unsigned long long) random()) << 32) ^ random();
}
#endif

/* Argument dir is the directory in which the files sts-XX are
 * located. */
static void init_files(const char *dir)
{
  if(files_initialized)
    return;

  int dir_len = strlen(dir);
  /* If necessary, add the trailing '/' or '\\' to the path. */
#if defined(__WIN32__)
  /* '\' instead of '/' */
  const char *sl = (dir_len > 0 && dir[dir_len-1] != '\\') ? "\\" : "";
#else
  const char *sl = (dir_len > 0 && dir[dir_len-1] != '/') ? "/" : "";
#endif

  for(int i = 0; i < FILE_COUNT; i++) {
    FILE *f;
    char filename[dir_len + FILENAME_LEN + 1];

    snprintf(filename, sizeof(filename), "%s%ssts19-%02d", dir, sl, i+1);
    if( (f = fopen(filename, "rb")) == NULL) {
      fprintf(stderr, "Unable to open file '%s'.\n", filename);
      abort();
    }
    files[i] = init_decompress(f);
  }

  files_initialized = 1;
}

const int * get_sts(long long idx)
{
  if(idx >= STS_COUNT) {
    fprintf(stderr, "get_sts(): invalid idx\n");
    abort();
  }
  if(!files_initialized) {
    fprintf(stderr, "get_sts(): files not opened\n");
    abort();
  }

  file_r *r = files[idx/STS_PER_FILE];
  idx       = idx % STS_PER_FILE;

  long setsize = get_set_size(r);
  open_set(r, idx/setsize);
  idx = idx % setsize;

  const int *res = NULL;
  for(long i = 0; i <= idx; i++)
    res = decompress(r);

  return res;
}


int main(int argc, char *argv[])
{
  const char *dir = argv[1];
  int       count = 10;

  if(argc <= 1 || argc > 3)
      goto usage;

  if(argc > 2) {
    count = strtol(argv[2], NULL, 10);
    if(count <= 0 || argc > 3)
      goto usage;
  }

  init_files(dir);

  for(int i = 0; i < count; i++) {
    long long idx = rnd() / (ULLONG_MAX / STS_COUNT);

    const int *sts = get_sts(idx);
    for(int j = 0; j < 3*STS_B; j++)
      fputc('a' + sts[j], stdout);
    fputc('\n', stdout);
  }

  return 0;

usage:
  fprintf(stderr, "usage: sampe path [count]\n");
  return 1;
}
