/*
 * stsc.c - command line utility for (de)compressing Steiner Triple Systems
 *
 * Copyright (c) 2008, Petteri Kaski, Patric Ostergard, Olli Pottonen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "compression.h"
#include <string.h>
#include <stdlib.h>

int main(int argc, char *argv[])
{
    int i, j;
    FILE *in, *out;

#define COMPRESS 0
#define DECOMPRESS 1
#define GET_SINGLE 2
    int mode;

    if( argc != 4 )
        goto usage;

    if( strcmp(argv[1], "-c") == 0 )
        mode=COMPRESS;
    else if( strcmp(argv[1], "-u") == 0 )
        mode=DECOMPRESS;
    else if( strcmp(argv[1], "-s") == 0 )
        mode=GET_SINGLE;
    else
        goto usage;

    if( strcmp(argv[2], "-") == 0 )
        in = stdin;
    else if( (in = fopen(argv[2], "rb")) == NULL ) {
        fprintf(stderr, "Unable to open file %s\n", argv[2]);
        return 1;
    }

    if( mode == COMPRESS || mode == DECOMPRESS ) {
        if( strcmp(argv[3], "-") == 0 )
            out = stdout;
        else if( (out = fopen(argv[3], "wb")) == NULL ) {
            fprintf(stderr, "Unable to open file %s\n", argv[3]);
            return 1;
        }
    } else // mode == GET_SINGLE
        out = stdout;


    /* output buffer */
    char csts[3*STS_B+2];
    csts[3*STS_B] = '\n';
    csts[3*STS_B+1] = '\0';

    if( mode == COMPRESS ) {
        /* input buffer */
        int t[3*STS_B];
        file_w *w;
        w = init_compress(out);

        for(;;) {
            /* Read STS */
            for( i=0 ; i<3*STS_B ; ++i)
                t[i] = fgetc(in) - 'a';
            fgetc(in); // Ignore newline
            if( feof(in) )
                break;
            /* Compress & output the STS */
            compress(t, w);
        }

        free_compress(w);
        return 0;

    } else if( mode == DECOMPRESS ) {
        file_r *r;
        r = init_decompress(in);

        fprintf(stderr, "Setsize: %u, number of systems: %u\n",
                get_set_size(r),
                get_sts_count(r));
        int stscount = get_sts_count(r);
        const int *sts;
        for( i=0 ; i<stscount ; ++i) {
            /* Decompress */
            sts = decompress(r);
            for( j=0 ; j<3*STS_B ; ++j)
                csts[j] = sts[j] + 'a';
            fputs(csts, out);
        }

        free_decompress(r);
        return 0;

    } else { // mode == GET_SINGLE
        file_r *r;
        r = init_decompress(in);

        int idx = atoi(argv[3])-1;
        if( idx < 0 || idx >= (int)get_sts_count(r) ) {
            fprintf(stderr, "Invalid parameter '%s'\n.", argv[3]);
                return 1;
        }

        const int *sts = NULL;
        int setsize = get_set_size(r);
        open_set(r, idx/setsize);
        for( i=0 ; i <= (int)(idx % setsize) ; ++i)
            sts = decompress(r);

        for( i=0 ; i<3*STS_B ; ++i)
            csts[i] = sts[i] + 'a';
        fputs(csts, stdout);

        free_decompress(r);
        return 0;
    }

  usage:
    fprintf(stderr, "usage: stsc (-u|-c) <inputfile> <outputfile> or\n"
                    "stsc -s <inputfile>\n");
    return 1;
}
